/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.hotmail.frojasg1.applications.fileencoder;

import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.StringFunctions;
import com.frojasg1.general.files.BinaryFile;
import com.frojasg1.general.keyboard.GenericKeyboard;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.OperationCancellation;
import com.frojasg1.general.progress.UpdatingProgress;
import com.hotmail.frojasg1.applications.fileencoder.configuration.EncryptingConfiguration;
import com.hotmail.frojasg1.applications.fileencoder.configuration.ListOfEncryptingConfigurations;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_1;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_2;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_3;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_4;
import com.hotmail.frojasg1.encrypting.encoderdecoder.EncoderDecoder;
import com.frojasg1.general.progress.GeneralUpdatingProgress;
import com.frojasg1.generic.GenericFunctions;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_5;
import com.hotmail.frojasg1.applications.fileencoder.fileencodertypes.FileEncoderType_6;
import com.hotmail.frojasg1.encrypting.randomnumbers.RandomSource;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.MessageDigest;
import java.nio.ByteBuffer;

public class FileEncoder
{
	protected static final int _minimumNumberOfBytesToReport = 4 * 1024*1024;
	
	public FileEncoder( )
	{
	}

	public static byte[] M_getBytesFromFile( String fileName ) throws FileEncoderException
	{
		return( M_getHeaderFromFile( fileName, -1 ) );
	}
	
	public static byte[] M_getHeaderFromFile( String fileName, int numberOfBytesToRead ) throws FileEncoderException
	{
		File file = new File(fileName);

		long numBytes = ( numberOfBytesToRead == -1 ? file.length() :
						( numberOfBytesToRead > file.length() ? file.length() : numberOfBytesToRead )
						);
		
		if( numBytes > 100000000 ) throw( new FileEncoderException( "File Too long: " + fileName ));
	    byte[] fileData = new byte[(int) numBytes];
	    DataInputStream dis = null;
	    try
	    {
	    	dis = new DataInputStream(new FileInputStream(file));
		    dis.readFully(fileData);
	    }
		catch( IOException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
		finally
		{
			try
			{
				if( dis != null ) dis.close();
			}
			catch( IOException ex )
			{
				ex.printStackTrace();
				throw( new FileEncoderException( ex.getMessage() ) );
			}
		}
	    return( fileData );
	}

	public static byte[] getFileData( String fileName, int numberOfBytesToDiscard ) throws FileEncoderException
	{
		byte[] fileData = null;

		File file = new File(fileName);

		byte[] header = new byte[numberOfBytesToDiscard];
		long numBytes = file.length() - numberOfBytesToDiscard;

		if( numBytes > 0 )
		{
			if( numBytes > 100000000 ) throw( new FileEncoderException( "File Too long: " + fileName ));
			fileData = new byte[(int) numBytes];
			DataInputStream dis = null;
			try
			{
				dis = new DataInputStream(new FileInputStream(file));
				dis.readFully(header);
				dis.readFully(fileData);
			}
			catch( IOException ex )
			{
				ex.printStackTrace();
				throw( new FileEncoderException( ex.getMessage() ) );
			}
			finally
			{
				try
				{
					if( dis != null ) dis.close();
				}
				catch( IOException ex )
				{
					ex.printStackTrace();
					throw( new FileEncoderException( ex.getMessage() ) );
				}
			}
		}
	    return( fileData );
	}

	public static byte[] M_getMD5( byte[] input ) throws FileEncoderException
	{
	      MessageDigest m = null;
	      try
	      {
	        m=MessageDigest.getInstance("MD5");
	      }
	      catch( Exception ex )
	      {
	        throw new FileEncoderException( "Excepcion intentando recoger el algoritmo MD5" );
	      }
	      m.update(input,0,input.length);
	      return( m.digest() );
	}
/*
	public static void M_writeEncodedFile( String newFileName, byte[] header, byte[] encodedFileData )
		throws FileEncoderException
	{
		FileOutputStream fos = null;
		try
		{
			fos = new FileOutputStream( new File( newFileName ) );
			fos.write(header);
			fos.flush();
			fos.write(encodedFileData);
			fos.flush();
		}
		catch( IOException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
		finally
		{
			try
			{
				if( fos != null ) fos.close();
			}
			catch( IOException ex )
			{
				ex.printStackTrace();
				throw( new FileEncoderException( ex.getMessage() ) );
			}
		}
	}
*/
    public static void M_writeFile( String newFileName, byte[] fileData ) throws FileEncoderException
    {
		FileOutputStream fos = null;
		try
		{
			fos = new FileOutputStream( new File( newFileName ) );
			fos.write(fileData);
			fos.flush();
		}
		catch( IOException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
		finally
		{
			try
			{
				if( fos != null ) fos.close();
			}
			catch( IOException ex )
			{
				ex.printStackTrace();
				throw( new FileEncoderException( ex.getMessage() ) );
			}
		}
    }

	public static String M_removejfeExtension( String fileName, String extension ) throws FileEncoderException
	{
		String result = null;
		String dotExtension = "." + extension;
		
		if( fileName.length() > dotExtension.length() )
		{
			if( fileName.substring( fileName.length()-dotExtension.length(), fileName.length() ).compareTo( dotExtension ) == 0 )
			{
				result = new String( fileName.substring( 0, fileName.length()-dotExtension.length() ));
			}
		}
		if( result == null ) throw( new FileEncoderException( "Bad file name for an encrypted file. It has to end in " + dotExtension ) );
		return( result );
	}

	public static void M_encodeFileFromBuffer( char[] password, FileEncoderParameters fep,
												byte[] dataToEncode,
												String fileName,
												UpdatingProgress up,
												OperationCancellation oc )  throws FileEncoderException, CancellationException
	{
		checkFreeSpace( fileName, dataToEncode.length );

		if( fep == null )
		{
			EncryptingConfiguration ec = ListOfEncryptingConfigurations.M_getInstance().
										M_getEncryptingConfigurationForAParticularFileSize(dataToEncode.length);

			if( ec != null )
			{
				fep = new FileEncoderParameters();
				
				ec.M_getParamListFromEncryptingConfiguration( fep );
			}
			else
			{
				System.out.println( "The file " + fep.M_getDecodedFileName() + " did not produce any encrypting configuration. May be the file does not exist." );
			}
		}

		FileEncoderType fet = null;
		fet = M_newFileEncoderType( fep.M_getFileEncoderType(), up );
		fet.M_encodeFileFromBuffer( password, fep, dataToEncode, fileName );
	}

	public static byte[] M_decodeFileToBuffer( char[] password, FileEncoderParameters fep,
														String fileName,
														UpdatingProgress up,
														OperationCancellation oc )  throws FileEncoderException, CancellationException
	{
		FileEncoderType fet = null;

		int fileEncoderType = -1;

		checkFreeSpace( fileName, -1 );

		byte[] header = M_getHeaderFromFile( fileName, 30 );
		byte[] headerHead = new byte[ FileEncoderParameters.STR_HEADER_HEAD.getBytes().length ];

		if( header.length > headerHead.length )
		{
			System.arraycopy(header, 0, headerHead, 0, headerHead.length );
			String str_headerHead = new String( headerHead );
			if( !str_headerHead.equals( FileEncoderParameters.STR_HEADER_HEAD ) )
			{
				throw( new FileEncoderException( "The supposed encrypted file has not got the appropriate format." ) );
			}

			fileEncoderType = header[ headerHead.length ];
		}
		else
		{
			throw( new FileEncoderException( "The supposed encrypted file has not got the appropriate format." ) );
		}

		fet = M_newFileEncoderType( fileEncoderType, up );

		byte[] decodedBuffer = fet.M_decodeFileToBuffer( password, fep, fileName );
		
		return( decodedBuffer );
	}

	/**
	 * 
	 * @param fileName
	 * @param length		If this parameter takes the value -1, then the free space is compared with the length of the file.
	 * @throws FileEncoderException 
	 */
	protected static void checkFreeSpace( String fileName, long length ) throws FileEncoderException
	{
		if( fileName != null )
		{
			File file = new File( fileName );
			
			length = ( ( length == -1 ) && file.exists() && file.isFile() ) ? file.length() : length;

			if( ( length != -1 ) && ( file.getParent() != null ) )
			{
				long usableSpace = file.getParentFile().getUsableSpace();
				long additionalSpaceNeeded = length - usableSpace;
				if( additionalSpaceNeeded > 0 )
				{
					throw( new FileEncoderException( "There is not enough free space in the disk to perform the operation." +
														"Approximate additional space needed: " + additionalSpaceNeeded + " bytes.\n" +
														"Disk of file: " + fileName ) );
				}
			}
		}
	}

	protected static void checkFreeSpace( String fileName ) throws FileEncoderException
	{
		checkFreeSpace( fileName, -1 );
	}

	public static FileEncoderParameters M_doActions( char[] password, String[] args, UpdatingProgress up ) throws FileEncoderException, CancellationException
	{
		int fileEncoderType = -1;

		FileEncoderParameters fep = new FileEncoderParameters( args, false, true );
		
		if( fep.M_getHasToUseFileSizeForEncryptingParams() && fep.M_isEncode() &&
			( fep.M_getDecodedFileName() != null ) )
		{
			EncryptingConfiguration ec = ListOfEncryptingConfigurations.M_getInstance().
										M_getEncrytingConfigurationFromFileSize(fep.M_getDecodedFileName());

			if( ec != null )
			{
				String argsFromFileSize[] = ec.M_getParamListFromEncryptingConfiguration();

				String args2[] = ArrayFunctions.joinStringArrays( argsFromFileSize, args );

				fep = new FileEncoderParameters( args2, false, true );
			}
			else
			{
				System.out.println( "The file " + fep.M_getDecodedFileName() + " did not produce any encrypting configuration. May be the file does not exist." );
			}
		}
		
		FileEncoderType fet = null;

		if( fep.M_isEncode() )
		{
			fileEncoderType = fep.M_getFileEncoderType();
			checkFreeSpace( fep.M_getDecodedFileName() );
		}
		else
		{
			String fileName = fep.M_getEncodedFileName();
			checkFreeSpace( fileName );

			byte[] header = M_getHeaderFromFile( fileName, 30 );
			byte[] headerHead = new byte[ FileEncoderParameters.STR_HEADER_HEAD.getBytes().length ];

			if( header.length > headerHead.length )
			{
				System.arraycopy(header, 0, headerHead, 0, headerHead.length );
				String str_headerHead = new String( headerHead );
				if( !str_headerHead.equals( FileEncoderParameters.STR_HEADER_HEAD ) )
				{
					throw( new FileEncoderException( "The supposed encrypted file has not got the appropriate format." ) );
				}

				fileEncoderType = header[ headerHead.length ];
			}
			else
			{
				throw( new FileEncoderException( "The supposed encrypted file has not got the appropriate format." ) );
			}
		}

		fet = M_newFileEncoderType( fileEncoderType, up );
		fet.M_doActions( password, fep );

		return( fet.M_getFileEncoderParameters() );
	}

	public static FileEncoderType M_newFileEncoderType( int fileEncoderType, UpdatingProgress up ) throws FileEncoderException
	{
		FileEncoderType result = null;

		switch( fileEncoderType )
		{
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_1: result = new FileEncoderType_1( up ); break;
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_2: result = new FileEncoderType_2( up ); break;
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_3: result = new FileEncoderType_3( up ); break;
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_4: result = new FileEncoderType_4( up ); break;
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_5: result = new FileEncoderType_5( up ); break;
			case FileEncoderParameters.INT_FILE_ENCODER_TYPE_6: result = new FileEncoderType_6( up ); break;

			default:
			{
				throw( new FileEncoderException( "Bad file encoder type: " + fileEncoderType ) );
			}
//			break;
		}

		return( result );
	}

	public static byte[] M_getMD5_fromFile_sliced( String fileName, int sliceSize,
													UpdatingProgress up )
					throws FileEncoderException, CancellationException
	{
		ByteBuffer bb = ByteBuffer.allocate( sliceSize );

		long length = getFileLength( fileName );
		sliceSize = getEffectiveSliceSize( sliceSize, length );

		int maxLeftToReport = 1;
		if( sliceSize > 0 )
			maxLeftToReport = (int) ( 25000000 / sliceSize + 1 );

		MessageDigest m = null;
		try
		{
			m=MessageDigest.getInstance("MD5");
		}
		catch( Exception ex )
		{
			throw new FileEncoderException( "Excepcion intentando recoger el algoritmo MD5" );
		}
		
		BinaryFile bf = null;

		try
		{
			if( up != null )
			{
				up.up_childStarts();
			}

			bf = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf.openForReading(fileName);

			long totalBytesRead = 0;
			int bytesRead = -1;
			int leftToReport = maxLeftToReport;

			while( ( bytesRead = bf.read( bb, totalBytesRead ) ) > 0 )
			{
				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer = bb.array();
				m.update(buffer,0,bytesRead);

				if( ( up != null ) && ( --leftToReport == 0 ) )
				{
					double progress = (double) totalBytesRead / length;
					up.up_updateProgressFromChild( progress );
					leftToReport = maxLeftToReport;
				}

				bb.clear();
			}

			if( up != null )
			{
				up.up_childEnds();
			}
		}
		catch( CancellationException ce )
		{
			throw( ce );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion calculating MD5 hash from sliced file" ) );
		}
		finally
		{
			try
			{
				if( bf != null ) bf.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}
		return( m.digest() );
	}
/*
	public static byte[] M_getMD5( byte[] buffer ) throws FileEncoderException
	{
		MessageDigest m = null;
		try
		{
			m=MessageDigest.getInstance("MD5");
		}
		catch( Exception ex )
		{
			throw new FileEncoderException( "Excepcion intentando recoger el algoritmo MD5" );
		}
		
		try
		{
			m.update(buffer,0,buffer.length);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion calculating MD5 hash from sliced file" ) );
		}
		return( m.digest() );
	}
*/

	public static void M_writeEncodedFile_sliced( String fileName, String newFileName,
													byte[] header, EncoderDecoder ed,
													int sliceSize, UpdatingProgress up )  throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		long decodedLength = getFileLength( fileName );
		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );
		ByteBuffer bb_out = ByteBuffer.wrap(header);

		BinaryFile bf_in = null;
		BinaryFile bf_out = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( decodedLength );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

//		gup.setDebug(true);
		try
		{
			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			int numBytesWritten = bf_out.write(bb_out);
			
			if( numBytesWritten != header.length ) throw( new FileEncoderException( "Error writing header to encoded file: " +
																					newFileName ) );
			
			long totalBytesRead = 0;
			int bytesRead = -1;

			while( ( bytesRead = readFully(bf_in, bb_in, totalBytesRead) ) > 0 )
			{
				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
				}

				byte[] buffer_out = ed.M_encode( buffer_in_wrapped, gup );
				bb_out = ByteBuffer.wrap( buffer_out );

				numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != bytesRead ) throw( new FileEncoderException( "Error writing file data to encoded file: " +
																					newFileName ) );

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Encrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when reading file " + fileName + " and writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}

			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}
	
	
	public static void M_writeDecodedFile_sliced( String fileName, String newFileName,
													int headerLength, EncoderDecoder ed,
													int sliceSize, UpdatingProgress up ) throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		long decodedLength = getFileLength( fileName ) - headerLength;
		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );
		ByteBuffer bb_out = null;

		File file = new File( fileName );
		long inFileSize = file.length();

		BinaryFile bf_in = null;
		BinaryFile bf_out = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( decodedLength );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

//		gup.setDebug(true);
		try
		{
			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			long totalBytesRead = headerLength;
			int bytesRead = -1;

//			gup.up_skip( headerLength );

			while( ( bytesRead = readFully(bf_in, bb_in, totalBytesRead) ) > 0 )
			{
				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
				}

				byte[] buffer_out = ed.M_decode( buffer_in_wrapped, gup );
				bb_out = ByteBuffer.wrap( buffer_out );

				int numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != bytesRead ) throw( new FileEncoderException( "Error writing file data to decoded file: " +
																					newFileName ) );

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Decrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when reading file " + fileName + " and writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}

			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}

	public static byte[] readFully( byte[] totalBuffer, int sliceSize, int totalBytesRead )
	{
		byte[] result = null;

		if( totalBuffer.length > totalBytesRead )
		{
			int bytesToRead = totalBuffer.length - totalBytesRead;
			if( ( sliceSize > 0 ) && ( bytesToRead > sliceSize ) )
				bytesToRead = sliceSize;

			result = new byte[ bytesToRead ];
			System.arraycopy( totalBuffer, totalBytesRead, result, 0, bytesToRead);
		}

		return( result );
	}

    public static int readFully(BinaryFile fc, ByteBuffer bb, long position ) throws FileEncoderException
	{
        int toRead = bb.remaining();
		int bytesRead = 0;
		int totalBytesRead = 0;

		try
		{
			while( (totalBytesRead < toRead) && (bytesRead != -1) )	// when the file ends (bytesRead==-1) is true
			{
				if( position >= 0 )
				{
					bytesRead = fc.read( bb, position );
					if( bytesRead > 0 ) position = position + bytesRead;
				}
				else
				{
					bytesRead = fc.read( bb );
				}

				if( bytesRead > 0 ) totalBytesRead = totalBytesRead + bytesRead;
			}
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( th.getMessage() ) );
		}

		return( totalBytesRead );
    }

	public static void encodeDataToFile( String newFileName,
											byte[] header, EncoderDecoder ed,
											int sliceSize, UpdatingProgress up,
											byte[] dataToEncode )  throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		ByteBuffer bb_out = ByteBuffer.wrap(header);

		long inFileSize = dataToEncode.length;

		sliceSize = getEffectiveSliceSize( sliceSize, inFileSize );

		BinaryFile bf_out = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( inFileSize );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(dataToEncode.length) );

//		gup.setDebug(true);
		try
		{
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			int numBytesWritten = bf_out.write(bb_out);
			
			if( numBytesWritten != header.length ) throw( new FileEncoderException( "Error writing header to encoded file: " +
																					newFileName ) );

			long totalBytesRead = 0;
			int bytesRead = -1;

			byte[] buffer_in = null;
			while( ( buffer_in = readFully(dataToEncode, sliceSize, (int) totalBytesRead) ) != null )
			{
				bytesRead = buffer_in.length;
				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
				}

				byte[] buffer_out = ed.M_encode( buffer_in_wrapped, gup );
				bb_out = ByteBuffer.wrap( buffer_out );
				
				numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != bytesRead ) throw( new FileEncoderException( "Error writing file data to encoded file: " +
																					newFileName ) );
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Encrypting file to ( " + newFileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}

	public static long getFileLength( String fileName )
	{
		long result = -1;
		File inputFile = new File( fileName );
		if( inputFile.exists() && inputFile.isFile() )
			result = inputFile.length();

		return( result );
	}

	protected static int getFileLengthLesserThan2GB( String fileName ) throws FileEncoderException
	{
		long result = getFileLength( fileName );
		
		if( result < 0 )
			throw( new FileEncoderException( "Problems reading file length for file: " + fileName ) );

		if ( result > 0x7FFFFFFFL )
			throw( new FileEncoderException( "File too big to be read completely in memory. File: " + fileName ) );

		return( (int) result );
	}

	protected static int getEffectiveSliceSize( int sliceSize, long fileLength ) throws FileEncoderException
	{
		int result = sliceSize;
		if( result == 0 )
		{
			if( result < 0 )
				throw( new FileEncoderException( "Problems reading file length" ) );

			if ( result > 0x7FFFFFFFL )
				throw( new FileEncoderException( "File too big to fit in one slice" ) );
			
			result = (int) fileLength;
		}
		else if( result < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than 0." ) );

		return( result );
	}

	public static byte[] decodeEncryptedFile( String fileName,
													int headerLength, EncoderDecoder ed,
													int sliceSize, UpdatingProgress up ) throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		long decodedLength = getFileLengthLesserThan2GB( fileName ) - headerLength;
		byte[] result = new byte[ (int) decodedLength ];

		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );

		BinaryFile bf_in = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( decodedLength );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

//		gup.setDebug(true);
		try
		{
			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);

			long totalBytesRead = headerLength;
			int bytesRead = -1;

			while( ( bytesRead = readFully(bf_in, bb_in, totalBytesRead) ) > 0 )
			{
				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
				}

				byte[] buffer_out = ed.M_decode( buffer_in_wrapped, gup );

				System.arraycopy(buffer_out, 0, result, (int) totalBytesRead - headerLength - bytesRead, buffer_in_wrapped.length);

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Decrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when decrypting file " + fileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();

		return( result );
	}

	public static void createFileWithZeroes( String fileName, long fileLength,
												UpdatingProgress up )  throws FileEncoderException, CancellationException
	{
		int sliceSize = 1024*1024;
		byte[] bufferOut = new byte[ sliceSize ];
		for( int ii=0; ii<sliceSize; ii++ )
			bufferOut[ii]=0;

		ByteBuffer bb_out = ByteBuffer.wrap(bufferOut);

		BinaryFile bf_out = null;

		if( up != null )
		{
			up.up_childStarts();
		}

		try
		{
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(fileName, truncateExisting);
			
			int numBytesWritten = 0;
			
			long totalBytesWritten = 0;
			int bytesToWrite = -1;

			while( ( bytesToWrite =
					( totalBytesWritten + sliceSize <= fileLength ?
						sliceSize :
						(int) ( fileLength - totalBytesWritten )  )  ) > 0 )
			{
				if( up != null )
				{
					double progress = (double) totalBytesWritten / fileLength;
					up.up_updateProgressFromChild( progress );
				}

				totalBytesWritten = totalBytesWritten + bytesToWrite;
				byte[] buffer_out_wrapped = null;

				if( bytesToWrite == sliceSize ) buffer_out_wrapped = bufferOut;
				else
				{
					buffer_out_wrapped = new byte[bytesToWrite];
					System.arraycopy(bufferOut, 0, buffer_out_wrapped, 0, bytesToWrite );
				}

				bb_out = ByteBuffer.wrap( buffer_out_wrapped );

				numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != bytesToWrite ) throw( new FileEncoderException( "Error writing file data zeroes to file: " +
																					fileName ) );
			}
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when writing zeroes to file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when reading file " + fileName + " and writing file " + fileName ) );
		}
		finally
		{
			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}


	public static void M_writeEncodedFile_randomlySliced( String fileName, String newFileName,
													byte[] header, int[] randomOrder,
													EncoderDecoder ed,
													int sliceSize, UpdatingProgress up )  throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		boolean isSliceSizeZero = ( sliceSize == 0 );
		
		long decodedLength = getFileLength( fileName );
		if( ( sliceSize != 0 ) &&
			( ( randomOrder == null ) ||
				( randomOrder.length != ( (decodedLength - 1) / sliceSize + 1 ) ) ) )
		{
			throw( new FileEncoderException( "randomOrder has not got the appropriate size" ) );
		}
		
		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );
		ByteBuffer bb_out = ByteBuffer.wrap(header);

		BinaryFile bf_in = null;
		BinaryFile bf_out = null;

		long totalLengthToEncode = ( isSliceSizeZero ? decodedLength : (long) sliceSize * randomOrder.length );

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( totalLengthToEncode );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

//		gup.setDebug(true);
		try
		{
			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			int numBytesWritten = bf_out.write(bb_out);

			if( numBytesWritten != header.length ) throw( new FileEncoderException( "Error writing header to encoded file: " +
																					newFileName ) );

			long totalBytesRead = 0;
			int bytesRead = -1;
			int hiest = ( isSliceSizeZero ? 1 : randomOrder.length );
			for( int ii=0; ii<hiest; ii++ )
			{
				long position = ( isSliceSizeZero ? 0 : (long) sliceSize * randomOrder[ii] );
				bytesRead = readFully(bf_in, bb_in, position);

				gup.up_prepareNextSlice(sliceSize);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[sliceSize];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
					for( int jj=bytesRead; jj<sliceSize; jj++ )
						buffer_in_wrapped[jj]=0;
				}

				byte[] buffer_out = ed.M_encode( buffer_in_wrapped, gup );
				bb_out = ByteBuffer.wrap( buffer_out );

				numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != buffer_out.length ) throw( new FileEncoderException( "Error writing file data to encoded file: " +
																					newFileName ) );

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Encrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when reading file " + fileName + " and writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}

			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}

	public static void M_writeDecodedFile_randomlySliced( String fileName, String newFileName,
													int headerLength,
													long decodedLength,
													int[] randomOrder,
													EncoderDecoder ed,
													int sliceSize, UpdatingProgress up ) throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		boolean isSliceSizeZero = ( sliceSize == 0 );

		long maxDecodedLength = getFileLength( fileName ) - headerLength;

		if( ( sliceSize != 0 ) &&
			( ( randomOrder == null ) ||
				( randomOrder.length != ( (decodedLength - 1) / sliceSize + 1 ) ) ) )
		{
			throw( new FileEncoderException( "randomOrder has not got the appropriate size" ) );
		}

		if( (decodedLength < (maxDecodedLength - sliceSize)) ||
			(decodedLength > maxDecodedLength ) )
		{
			throw( new FileEncoderException( "Length of encodedFile does not match the stored length of decoded file." ) );
		}

		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );
		ByteBuffer bb_out = null;

		File file = new File( fileName );
		long inFileSize = file.length();

		BinaryFile bf_in = null;
		BinaryFile bf_out = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( decodedLength * 1.01D );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

		gup.up_prepareNextSlice( decodedLength * 0.01D );
//		gup.setDebug(true);
		try
		{
			createFileWithZeroes( newFileName, decodedLength, gup );

			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			long totalBytesRead = headerLength;
			int bytesRead = -1;
			
//			gup.up_skip( headerLength );

			int hiest = ( isSliceSizeZero ? 1 : randomOrder.length );
			for( int ii=0; ii<hiest; ii++ )
			{
				bytesRead = readFully(bf_in, bb_in, headerLength + (long) sliceSize * ii );// all slices are complete, even the last one.

				int index = ( isSliceSizeZero ? 0 : randomOrder[ii] );
				long position = (long) sliceSize * index;

				if( !isSliceSizeZero && ( index == (randomOrder.length-1) ) )	// the last slice may be not complete.
				{
					bytesRead = (int) (decodedLength % sliceSize);
					if( bytesRead == 0 )
						bytesRead = sliceSize;
				}

				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();

				byte[] buffer_out = ed.M_decode( buffer_in, gup );

				byte[] buffer_out_wrapped = null;

				// if it is a whole slice ...
				if( bytesRead == sliceSize )
					buffer_out_wrapped = buffer_out;
				else
				{	// if it is a fraction of a slice ...
					buffer_out_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_out, 0, buffer_out_wrapped, 0, bytesRead );
				}

				bb_out = ByteBuffer.wrap( buffer_out_wrapped );

				int numBytesWritten = bf_out.write(bb_out, position );

				if( numBytesWritten != bytesRead ) throw( new FileEncoderException( "Error writing file data to decoded file: " +
																					newFileName ) );

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Decrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when reading file " + fileName + " and writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}

			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}


	public static void encodeDataToFile_randomlySliced( String newFileName,
											byte[] header, EncoderDecoder ed,
											int sliceSize, UpdatingProgress up,
											byte[] dataToEncode,
											int[] randomOrder )  throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		boolean isSliceSizeZero = ( sliceSize == 0 );

		ByteBuffer bb_out = ByteBuffer.wrap(header);

		long inFileSize = dataToEncode.length;

		if( ( sliceSize != 0 ) &&
			( ( randomOrder == null ) ||
				( randomOrder.length != ( (inFileSize - 1) / sliceSize + 1 ) ) ) )
		{
			throw( new FileEncoderException( "randomOrder has not got the appropriate size" ) );
		}

		sliceSize = getEffectiveSliceSize( sliceSize, inFileSize );

		BinaryFile bf_out = null;

		int totalLengthToEncode = ( isSliceSizeZero ? (int) inFileSize : sliceSize * randomOrder.length );
		
		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( totalLengthToEncode );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(inFileSize) );

//		gup.setDebug(true);
		try
		{
			bf_out = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			boolean truncateExisting = true;
			bf_out.openForReadingAndWriting(newFileName, truncateExisting);

			int numBytesWritten = bf_out.write(bb_out);

			if( numBytesWritten != header.length ) throw( new FileEncoderException( "Error writing header to encoded file: " +
																					newFileName ) );

			long totalBytesRead = 0;
			int bytesRead = -1;

			byte[] buffer_in = null;

			int hiest = ( isSliceSizeZero ? 1 : randomOrder.length );
			for( int ii=0; ii<hiest; ii++ )
			{
				long position = ( isSliceSizeZero ? 0 : (long) sliceSize * randomOrder[ii] );
				buffer_in = readFully(dataToEncode, sliceSize, (int) position);

				bytesRead = buffer_in.length;
				gup.up_prepareNextSlice(sliceSize);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in_wrapped = null;

				if( bytesRead == sliceSize ) buffer_in_wrapped = buffer_in;
				else
				{
					buffer_in_wrapped = new byte[ sliceSize ];
					System.arraycopy(buffer_in, 0, buffer_in_wrapped, 0, bytesRead );
					for( int jj=bytesRead; jj<sliceSize; jj++ )
						buffer_in_wrapped[jj]=0;
					bytesRead = sliceSize;
				}

				byte[] buffer_out = ed.M_encode( buffer_in_wrapped, gup );
				bb_out = ByteBuffer.wrap( buffer_out );

				numBytesWritten = bf_out.write(bb_out );
				if( numBytesWritten != bytesRead ) throw( new FileEncoderException( "Error writing file data to encoded file: " +
																					newFileName ) );
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Encrypting file to ( " + newFileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when writing file " + newFileName ) );
		}
		finally
		{
			try
			{
				if( bf_out != null ) bf_out.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();
	}

	public static byte[] decodeEncryptedFile_randomlySliced( String fileName,
												int headerLength,
												int decodedLength, int[] randomOrder,
												EncoderDecoder ed,
												int sliceSize, UpdatingProgress up ) throws FileEncoderException, CancellationException
	{
		if( sliceSize < 0 )
			throw( new FileEncoderException( "Slice size must be greater or equal than zero" ) );

		boolean isSliceSizeZero = ( sliceSize == 0 );

		int maxDecodedLength = getFileLengthLesserThan2GB( fileName ) - headerLength;

		if( ( sliceSize != 0 ) &&
			( ( randomOrder == null ) ||
				( randomOrder.length != ( (decodedLength - 1) / sliceSize + 1 ) ) ) )
		{
			throw( new FileEncoderException( "randomOrder has not got the appropriate size" ) );
		}

		if( (decodedLength < (maxDecodedLength - sliceSize)) ||
			(decodedLength > maxDecodedLength ) )
		{
			throw( new FileEncoderException( "Length of encodedFile does not match the stored length of decoded file." ) );
		}

		byte[] result = new byte[ (int) decodedLength ];

		sliceSize = getEffectiveSliceSize( sliceSize, decodedLength );

		ByteBuffer bb_in = ByteBuffer.allocate( sliceSize );

		BinaryFile bf_in = null;

		GeneralUpdatingProgress gup = null;
		gup = new GeneralUpdatingProgress();
		gup.up_setParentUpdatingProgress( up );
		gup.up_reset( decodedLength * 1.01D );
		gup.up_prepareNextSlice( decodedLength * 0.01D );
		gup.up_setMinimumPercentageToReport( getMinimumFractionToReport(decodedLength) );

//		gup.setDebug(true);
		try
		{
			bf_in = GenericFunctions.instance().getFileFacilities().createBinaryFile();
			bf_in.openForReading(fileName);

			long totalBytesRead = headerLength;
			int bytesRead = -1;

			int hiest = ( isSliceSizeZero ? 1 : randomOrder.length );
			for( int ii=0; ii<hiest; ii++ )
			{
				bytesRead = readFully(bf_in, bb_in, headerLength + sliceSize * ii);

				int index = ( isSliceSizeZero ? 0 : randomOrder[ii] );
				int position = sliceSize * index;

				if( !isSliceSizeZero && ( index == (randomOrder.length-1) ) )
				{
					bytesRead = (int) (decodedLength % sliceSize);
					if( bytesRead == 0 )
						bytesRead = sliceSize;
				}

				gup.up_prepareNextSlice(bytesRead);

				totalBytesRead = totalBytesRead + bytesRead;
				byte[] buffer_in = bb_in.array();
				byte[] buffer_out = ed.M_decode( buffer_in, gup );

				byte[] buffer_out_wrapped = null;

				// if it is a whole slice ...
				if( bytesRead == sliceSize )
					buffer_out_wrapped = buffer_out;
				else
				{	// if it is a fraction of a slice ...
					buffer_out_wrapped = new byte[bytesRead];
					System.arraycopy(buffer_out, 0, buffer_out_wrapped, 0, bytesRead );
				}

				System.arraycopy(buffer_out, 0, result, position, buffer_out_wrapped.length);

				bb_in.clear();
			}

			gup.up_performEnd();
		}
		catch( CancellationException ce )
		{
			throw( new CancellationException( ce.getMessage() + " when Decrypting file ( " + fileName + " )" ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			throw( new FileEncoderException( "Excepcion when decrypting file " + fileName ) );
		}
		finally
		{
			try
			{
				if( bf_in != null ) bf_in.close();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
		}

		if( up != null )
			up.up_childEnds();

		return( result );
	}

	public static double getMinimumFractionToReport( double minimumFractionToReport,
														long fileSize )
	{
		return( DoubleFunctions.min(minimumFractionToReport,
									( fileSize != 0 ?
									  ( (double) _minimumNumberOfBytesToReport / fileSize ) :
									1.0D ) ) );
	}

	public static double getMinimumFractionToReport( long fileSize )
	{
		return( getMinimumFractionToReport( 0.001D, fileSize ) );
	}

	protected static char[] getPasswordFromKeyboard()
	{
		char[] result = null;

		try
		{
			GenericKeyboard gk = GenericFunctions.instance().getKeyboardFacilities();

			while( result == null )
			{
				System.out.println( "Type password and press Return:" );
				char[] tmp1 = gk.readPasswordFromKeyboard();

				System.out.println( "Type the password again and press Return:" );
				char[] tmp2 = gk.readPasswordFromKeyboard();

				if( StringFunctions.M_compare( tmp1, tmp2) == 0 )
					result = tmp1;
				else
				{
					System.out.println( "The passwords do not match. Do you want to try again (y/n) + return" );
					String yn = gk.readLineFromKeyboard();
					if( ( yn == null ) || !yn.toLowerCase().equals( "y" ) )
						break;
				}
			}
		}
		catch( Throwable th )
		{
			result = null;
			th.printStackTrace();
		}

		return( result );
	}

	/**
	 * Common main for command line invocation.
	 * This main cannot be invoked directly, because FileEncoderSDF instance must be created previously.
	 * FileEncoderSDF instance is system depending.
	 * 
	 * @param args 
	 */
	public static void main( String[] args )
	{
		RandomSource.instance();

		int returnValue = 0;

		try
		{
			UpdatingProgress up = new UpdatingProgress()
							{
								public void up_childStarts()
								{
									System.out.println( "Beginning" );
								}

								public void up_childEnds()
								{
									System.out.println( "End" );
								}

								@Override
								public void up_updateProgressFromChild(double completedOverOne) {
									System.out.println( String.format( "Progress: %,.4f %%", (completedOverOne * 100) ) );
								}

								@Override
								public OperationCancellation up_getOperationCancellation() {
									return( null );
								}

								@Override
								public void up_setDebug(boolean debug) {
								}
							};

			GeneralUpdatingProgress gup = new GeneralUpdatingProgress();
			gup.up_setParentUpdatingProgress( up );
			gup.up_reset( 1.0D );
			gup.up_prepareNextSlice( 1.0D );
			gup.up_setMinimumPercentageToReport( 0.001D );

			FileEncoderParameters fep = new FileEncoderParameters( args, false, true );

			char[] password = null;

			if( fep.M_getPassword() == null )
				password = getPasswordFromKeyboard();

			M_doActions(	password,
							args,
							up );

			gup.up_performEnd();

			returnValue = 0;
		}
		catch( Throwable ex )
		{
			ex.printStackTrace();
			returnValue = 1;
		}
		finally
		{
			RandomSource.instance().setMustStop(true);

			while( RandomSource.instance().isAlive() )
			{
				try
				{
					Thread.sleep( 100 );
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
			}
		}

		System.exit( returnValue );
/*
		try
		{
			String fileName = "J:\\N\\temp2\\Secuencias_pseudoaleatorias.pdf";
//			fileName = "C:\\Users\\Usuario\\frojasg1.apps\\FileEncoder\\general\\GlobalConfiguration.properties";
			byte[] fileData = M_getBytesFromFile( fileName );
			byte[] md5 =  M_getMD5( fileData );

			byte[] md5_2 = M_getMD5_fromFile_sliced( fileName, 65536 );

			String md5_1_string = HexadecimalFunctions.M_convertByteArrayToHexadecimalString(md5);
			String md5_2_string = HexadecimalFunctions.M_convertByteArrayToHexadecimalString(md5_2);

			System.out.println( "md5_1: " + md5_1_string );
			System.out.println( "md5_2: " + md5_2_string );

			System.out.println( "son iguales: " + (boolean) ( md5_1_string.equals( md5_2_string ) ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
*/
	}

//	protected boolean a_isEncode = false;
//	protected boolean a_hasToEraseOriginalFile = false;
//	protected String a_password = null;
//	protected String[] a_files = null;

//	protected static FileEncoder a_fe = null;

//	protected static RandomSource a_rs = null;
	
//	protected ReordererEncoder a_encoder = null;
}
