/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.dialogs.implementation;

import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.desktop.dialogs.DesktopViewComponent;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.dialogs.FilterForFile;
import com.frojasg1.general.dialogs.ViewComponent;
import java.awt.Component;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 *
 * @author Usuario
 */
public class DesktopDialogsWrapper implements DialogsWrapper
{
	BaseApplicationConfigurationInterface _conf = null;

	protected static DesktopDialogsWrapper _instance = null;

	public static DesktopDialogsWrapper createInstance( BaseApplicationConfigurationInterface conf )
	{
		_instance = new DesktopDialogsWrapper( conf );

		return( _instance );
	}

	public static DesktopDialogsWrapper instance()
	{
		return( _instance );
	}

	protected DesktopDialogsWrapper( BaseApplicationConfigurationInterface conf )
	{
		_conf = conf;
	}

	@Override
	public void showMessageDialog(ViewComponent parent, Object message, String title, int messageType)
	{
		StaticDesktopDialogsWrapper.showMessageDialog( getComponent( parent ),
														message, title,
														translateMessageType( messageType ),
														_conf);
	}

	@Override
	public int showOptionDialog(ViewComponent parent, Object message,
								String title, int optionType, int messageType,
								Object[] options, Object initialValue)
	{
		return(
				StaticDesktopDialogsWrapper.showOptionDialog( getComponent( parent ),
																message, title,
																translateOptionType( optionType ),
																translateMessageType( messageType ),
																null, options,
																initialValue,
																_conf )
			);
	}

	@Override
	public String showFileChooserDialog(ViewComponent parent, int open_save_dialog,
										List<FilterForFile> fileFilterList,
										String defaultFileName)
	{
		List<FileFilter> ffl = getFileFilterList( fileFilterList );

		int osd = translateOpenSaveDialog( open_save_dialog );
		
		String file = StaticDesktopDialogsWrapper.showFileChooserDialog( getComponent( parent ),
													osd,
													_conf,
													ffl,
													defaultFileName );

		return( file );
	}

	protected List<FileFilter> getFileFilterList( List<FilterForFile> fileFilterList )
	{
		List<FileFilter> ffl = null;
		
		if( fileFilterList != null )
		{
			ffl = new ArrayList<FileFilter>();
			Iterator<FilterForFile> it = fileFilterList.iterator();
			while( it.hasNext() )
			{
				FilterForFile fff = it.next();

				FileNameExtensionFilter fnef = new FileNameExtensionFilter( fff.getDescription(),
																				fff.getExtension() );

				ffl.add( fnef );
			}
		}
		
		return( ffl );
	}
	
	protected Component getComponent( ViewComponent vc )
	{
		Component comp = null;
		
		if( vc instanceof DesktopViewComponent )
		{
			DesktopViewComponent dvc = (DesktopViewComponent) vc;

			comp = dvc.getComponent();
		}
		
		return( comp );
	}

	protected int translateOpenSaveDialog( int open_save_dialog )
	{
		int result = StaticDesktopDialogsWrapper.OPEN;

		switch( open_save_dialog )
		{
			case OPEN:		result = StaticDesktopDialogsWrapper.OPEN;		break;
			case SAVE:		result = StaticDesktopDialogsWrapper.SAVE;		break;
		}
		
		return( result );
	}

	protected int translateMessageType( int messageType )
	{
		int result = JOptionPane.INFORMATION_MESSAGE;

		switch( messageType )
		{
			case ERROR_MESSAGE:				result = JOptionPane.ERROR_MESSAGE;				break;
			case INFORMATION_MESSAGE:		result = JOptionPane.INFORMATION_MESSAGE;		break;
			case PLAIN_MESSAGE:				result = JOptionPane.PLAIN_MESSAGE;				break;
			case QUESTION_MESSAGE:			result = JOptionPane.QUESTION_MESSAGE;			break;
			case WARNING_MESSAGE:			result = JOptionPane.WARNING_MESSAGE;			break;
		}
		
		return( result );
	}

	protected int translateOptionType( int optionType )
	{
		int result = JOptionPane.INFORMATION_MESSAGE;

		switch( optionType )
		{
			case CANCEL_OPTION:				result = JOptionPane.CANCEL_OPTION;				break;
			case CLOSED_OPTION:				result = JOptionPane.CLOSED_OPTION;				break;
			case DEFAULT_OPTION:			result = JOptionPane.DEFAULT_OPTION;			break;
			case OK_CANCEL_OPTION:			result = JOptionPane.OK_CANCEL_OPTION;			break;
			case OK_OPTION:					result = JOptionPane.OK_OPTION;					break;
			case YES_NO_CANCEL_OPTION:		result = JOptionPane.YES_NO_CANCEL_OPTION;		break;
			case YES_NO_OPTION:				result = JOptionPane.YES_NO_OPTION;				break;
			case YES_OPTION:				result = JOptionPane.YES_OPTION;				break;
		}

		return( result );
	}

	public ViewComponent createViewComponent( Component comp )
	{
		return( new SimpleViewComponent( comp ) );
	}

	public static class SimpleViewComponent implements DesktopViewComponent
	{
		protected Component _comp = null;

		public SimpleViewComponent( Component comp )
		{
			_comp = comp;
		}

		@Override
		public Component getComponent()
		{
			return( _comp );
		}
	};

}
