/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.splash;

import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.ResourceFunctions;
import com.frojasg1.general.desktop.dialogs.implementation.StaticDesktopDialogsWrapper;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.general.number.IntegerFunctions;
import com.hotmail.frojasg1.applications.common.components.internationalization.InternException;
import com.hotmail.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import javax.swing.SwingUtilities;
import javax.swing.plaf.basic.BasicProgressBarUI;

/**
 *
 * @author Usuario
 */
public class GenericBasicSplash extends InternationalizedJFrame implements GenericSplashInterface, MouseMotionListener, MouseListener
{
	protected BufferedImage _image = null;
	protected ImageJPanel _imageJPanel = null;

	protected Point _positionOnScreenWhenMousePressed = null;
	protected Point _positionOnScreenOfSplashWhenMousePressed = null;

	protected String _currentActivityLabel = null;
	
	/**
	 * Creates new form Splash
	 */
	public GenericBasicSplash( BaseApplicationConfigurationInterface applicationConfiguration )
	{
		super();
//		setImage( image );
		setUndecorated( true );

		this.setAppliConf(applicationConfiguration);

		initComponents();

		initOwnComponents_Parent();
	}

	protected final void initOwnComponents_Parent()
	{
		// http://stackoverflow.com/questions/3480125/setting-the-colors-of-a-jprogressbar-text
		jPB_progress.setUI( new BasicProgressBarUI() {
								protected Color getSelectionBackground() { return Color.black; }
								protected Color getSelectionForeground() { return Color.black; }
							  });
		
		setPreferredSize( jPanel1.getSize() );
		setSize( jPanel1.getSize() );
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jPanel1 = new javax.swing.JPanel();
        jPanelImage = new javax.swing.JPanel();
        jPB_progress = new javax.swing.JProgressBar();
        jL_currentActivity = new javax.swing.JLabel();

        setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
        setResizable(false);
        getContentPane().setLayout(null);

        jPanel1.setBorder(javax.swing.BorderFactory.createTitledBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)), "", javax.swing.border.TitledBorder.DEFAULT_JUSTIFICATION, javax.swing.border.TitledBorder.DEFAULT_POSITION, null, new java.awt.Color(0, 0, 0)));
        jPanel1.setLayout(null);

        jPanelImage.setBorder(javax.swing.BorderFactory.createLineBorder(new java.awt.Color(0, 0, 0)));

        javax.swing.GroupLayout jPanelImageLayout = new javax.swing.GroupLayout(jPanelImage);
        jPanelImage.setLayout(jPanelImageLayout);
        jPanelImageLayout.setHorizontalGroup(
            jPanelImageLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 108, Short.MAX_VALUE)
        );
        jPanelImageLayout.setVerticalGroup(
            jPanelImageLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 108, Short.MAX_VALUE)
        );

        jPanel1.add(jPanelImage);
        jPanelImage.setBounds(120, 20, 110, 110);

        jPB_progress.setForeground(new java.awt.Color(255, 0, 0));
        jPB_progress.setStringPainted(true);
        jPanel1.add(jPB_progress);
        jPB_progress.setBounds(40, 150, 280, 17);

        jL_currentActivity.setPreferredSize(new java.awt.Dimension(280, 10));
        jPanel1.add(jL_currentActivity);
        jL_currentActivity.setBounds(40, 170, 280, 20);

        getContentPane().add(jPanel1);
        jPanel1.setBounds(0, 0, 350, 210);

        pack();
    }// </editor-fold>//GEN-END:initComponents

	/**
	 * @param args the command line arguments
	 */
//	public static void main(String args[]) {
		/* Set the Nimbus look and feel */
        //<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
		 * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html 
		 */
	/*
		try {
			for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
				if ("Nimbus".equals(info.getName())) {
					javax.swing.UIManager.setLookAndFeel(info.getClassName());
					break;
				}
			}
		} catch (ClassNotFoundException ex) {
			java.util.logging.Logger.getLogger(GenericBasicSplash.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (InstantiationException ex) {
			java.util.logging.Logger.getLogger(GenericBasicSplash.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (IllegalAccessException ex) {
			java.util.logging.Logger.getLogger(GenericBasicSplash.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		} catch (javax.swing.UnsupportedLookAndFeelException ex) {
			java.util.logging.Logger.getLogger(GenericBasicSplash.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
		}
	*/
        //</editor-fold>
        //</editor-fold>

		/* Create and display the form */
//		java.awt.EventQueue.invokeLater(new Runnable() {
//			public void run() {
//				new GenericBasicSplash().setVisible(true);
//			}
//		});
//	}

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel jL_currentActivity;
    private javax.swing.JProgressBar jPB_progress;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JPanel jPanelImage;
    // End of variables declaration//GEN-END:variables

	protected void setImage( BufferedImage image )
	{
		if( ( image == null ) || ( image.getHeight() == 0 ) )
		{
			_image = new BufferedImage( 245, 120, BufferedImage.TYPE_INT_RGB );
			Graphics gc = _image.getGraphics();
			gc.setColor( Color.WHITE );
			gc.fillRect( 0, 0, 245, 120 );
		}
		else
			_image = image;

		setImageChanges();
	}

	protected void setImageChanges()
	{
		_imageJPanel = new ImageJPanel( _image );
		_imageJPanel.setFrameThick( 3 );

		jPanelImage.add( _imageJPanel );

		int width = (int) ( jPanelImage.getHeight() * _image.getWidth() / _image.getHeight() );
		int height = (int) jPanelImage.getHeight();

		Rectangle jpiBounds = jPanelImage.getBounds();
		jPanelImage.setBounds( (int) ( jpiBounds.getX() + ( jpiBounds.getWidth() - width ) / 2 ),
								(int) ( jpiBounds.getY() + ( jpiBounds.getHeight() - height ) / 2 ),
								width, height );
		_imageJPanel.setBounds( 0, 0, width, height );
		_imageJPanel.setPreferredSize( new Dimension( width, height ) );
	}

	@Override
	public void setCurrentActivity(String currentActivity)
	{
		jL_currentActivity.setText( currentActivity );
//		jL_currentActivity.update(jL_currentActivity.getGraphics());
	    jL_currentActivity.paintImmediately(jL_currentActivity.getVisibleRect());

		_currentActivityLabel = null;
	}

	@Override
	public void setCurrentActivityFromLabel( String label )
	{
		if( label != null )
		{
			setCurrentActivity( this.getInternationalString(label) );
			_currentActivityLabel = label;
		}
	}

	@Override
	public void setProgress(int progress)
	{
		progress = IntegerFunctions.min( jPB_progress.getMaximum(),
										IntegerFunctions.max( progress, jPB_progress.getMinimum() ) );

		jPB_progress.setValue( progress );
//		jPB_progress.update(jPB_progress.getGraphics());
	}

	@Override
	public void activateProgress( boolean activate )
	{
		jPB_progress.setVisible( activate );
		jL_currentActivity.setVisible( activate );
	}

	public void setFrameThick( int value )
	{
		_imageJPanel.setFrameThick(value);
	}

	public void setFontForTextLines( Font font )
	{
		_imageJPanel.setFontForTextLines(font);
	}

	public void setForegroundColorForTextLines( Color color )
	{
		_imageJPanel.setForegroundColorForTextLines(color);
	}

	public void addTextLine( ImageJPanel.TextLine textLine )
	{
		_imageJPanel.addTextLine(textLine);
	}

	public void clearTextLines()
	{
		_imageJPanel.clearTextLines();
	}

	public Dimension getSizeOfImageJPanel()
	{
		return( _imageJPanel.getSize() );
	}

	protected BufferedImage getImage( String pathInJarForApplicationImage )
	{
		BufferedImage result = null;
		try
		{
			result = ResourceFunctions.loadResourceImage(pathInJarForApplicationImage);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

		return( result );
	}

	public void setCenteredLocation()
	{
		setLocation(StaticDesktopDialogsWrapper.getCenteredLocationForComponent( this ) );
	}

	@Override
	public void mouseDragged(MouseEvent me)
	{
		if( _positionOnScreenWhenMousePressed != null )
		{
			Point pos = me.getLocationOnScreen();
			
			setLocation( (int) ( _positionOnScreenOfSplashWhenMousePressed.getX() + pos.getX() - _positionOnScreenWhenMousePressed.getX() ),
						(int) ( _positionOnScreenOfSplashWhenMousePressed.getY() + pos.getY() - _positionOnScreenWhenMousePressed.getY() ) );
		}
	}

	@Override
	public void mouseMoved(MouseEvent me)
	{
	}

	@Override
	public void mouseClicked(MouseEvent me)
	{
	}

	@Override
	public void mousePressed(MouseEvent me)
	{
		_positionOnScreenWhenMousePressed = me.getLocationOnScreen();
		_positionOnScreenOfSplashWhenMousePressed = getLocation();
	}

	@Override
	public void mouseReleased(MouseEvent me)
	{
	}

	@Override
	public void mouseEntered(MouseEvent me)
	{
	}

	@Override
	public void mouseExited(MouseEvent me)
	{
	}
	
	protected void setListenersRecursive( Component comp )
	{
		addMouseMotionListener(this);
		addMouseListener(this);

		if( comp instanceof Container )
		{
			Container cont = (Container) comp;
			
			for( int ii=0; ii<cont.getComponentCount(); ii++ )
				setListenersRecursive( cont.getComponent(ii) );
		}
	}

	@Override
	public void changeLanguage( String language ) throws ConfigurationException, InternException
	{
		super.changeLanguage( language );
		setCurrentActivityFromLabel( _currentActivityLabel );
	}

}
